@@ -3,17 +3,20 @@ import queue
 import threading
 import json
 import time
+import re
 from collections import defaultdict
 from enum import Enum
 from contextlib import suppress, contextmanager
 from copy import copy
-from typing import Callable, Mapping, Optional, Union
+from typing import Any, Callable, Mapping, Optional, Union, Pattern
 from ssl import SSLContext
 import abc
 
 from werkzeug.http import parse_authorization_header
 from werkzeug.serving import make_server
-from werkzeug.wrappers import Request, Response
+from werkzeug.wrappers import Request
+from werkzeug.wrappers import Response
+
 import werkzeug.urls
 from werkzeug.datastructures import MultiDict
 
@@ -21,6 +24,14 @@ URI_DEFAULT = ""
 METHOD_ALL = "__ALL"
 
 
+class Undefined:
+    def __repr__(self):
+        return "<UNDEFINED>"
+
+
+UNDEFINED = Undefined()
+
+
 class Error(Exception):
     """
     Base class for all exception defined in this package.
@@ -133,6 +144,7 @@ class QueryMatcher(abc.ABC):
     get_comparing_values should return a 2-element tuple whose elements will
     be compared.
     """
+
     def match(self, request_query_string: bytes) -> bool:
         values = self.get_comparing_values(request_query_string)
         return values[0] == values[1]
@@ -141,10 +153,12 @@ class QueryMatcher(abc.ABC):
     def get_comparing_values(self, request_query_string: bytes) -> tuple:
         pass
 
+
 class StringQueryMatcher(QueryMatcher):
     """
     Matches a query for a string or bytes specified
     """
+
     def __init__(self, query_string: Union[bytes, str]):
         """
         :param query_string: the query string will be compared to this string or bytes.
@@ -152,19 +166,18 @@ class StringQueryMatcher(QueryMatcher):
             The query must not start with '?' but will be exactly (byte-by-byte) equal
             the actual query string of the incoming request.
         """
-        if query_string is not None and not isinstance(query_string, (str, bytes)):
+        if not isinstance(query_string, (str, bytes)):
             raise TypeError("query_string must be a string, or a bytes-like object")
 
         self.query_string = query_string
 
     def get_comparing_values(self, request_query_string: bytes) -> tuple:
-        if self.query_string is not None:
-            if isinstance(self.query_string, str):
-                query_string = self.query_string.encode()
-            elif isinstance(self.query_string, bytes):
-                query_string = self.query_string
-            else:
-                raise TypeError("query_string must be a string, or a bytes-like object")
+        if isinstance(self.query_string, str):
+            query_string = self.query_string.encode()
+        elif isinstance(self.query_string, bytes):
+            query_string = self.query_string
+        else:
+            raise TypeError("query_string must be a string, or a bytes-like object")
 
         return (request_query_string, query_string)
 
@@ -174,7 +187,7 @@ class MappingQueryMatcher(QueryMatcher):
     Matches a query string to a dictionary or MultiDict specified
     """
 
-    def __init__(self, query_dict: [Mapping, MultiDict]):
+    def __init__(self, query_dict: Union[Mapping, MultiDict]):
         """
         :param query_dict: if dictionary (Mapping) is specified, it will be used as a
             key-value mapping where both key and value should be string. If there are multiple
@@ -196,6 +209,7 @@ class BooleanQueryMatcher(QueryMatcher):
     """
     Matches the query depending on the boolean value
     """
+
     def __init__(self, result: bool):
         """
         :param result: if this parameter is true, the query match will be always
@@ -226,13 +240,27 @@ def _create_query_matcher(query_string: Union[None, QueryMatcher, str, bytes, Ma
     raise TypeError("Unable to cast this type to QueryMatcher: {!r}".format(type(query_string)))
 
 
+class URIPattern(abc.ABC):
+    @abc.abstractmethod
+    def match(self, uri: str) -> bool:
+        """
+            Matches the provided URI.
+
+            :param uri: URI of the request. This is an absolute path startting
+                with "/" and does not contain the query part.
+            :return: True if there's a match, False otherwise
+            """
+        pass
+
+
 class RequestMatcher:
     """
     Matcher object for the incoming request.
 
     It defines various parameters to match the incoming request.
 
-    :param uri: URI of the request. This must be an absolute path starting with ``/``.
+    :param uri: URI of the request. This must be an absolute path starting with ``/``, a
+        :py:class:`URIPattern` object, or a regular expression compiled by :py:func:`re.compile`.
     :param method: HTTP method of the request. If not specified (or `METHOD_ALL`
         specified), all HTTP requests will match.
     :param data: payload of the HTTP request. This could be a string (utf-8 encoded
@@ -249,18 +277,23 @@ class RequestMatcher:
 
     def __init__(
             self,
-            uri: str,
+            uri: Union[str, URIPattern, Pattern[str]],
             method: str = METHOD_ALL,
             data: Union[str, bytes, None] = None,
             data_encoding: str = "utf-8",
             headers: Optional[Mapping[str, str]] = None,
             query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-            header_value_matcher: Optional[HeaderValueMatcher] = None):
+            header_value_matcher: Optional[HeaderValueMatcher] = None,
+            json: Any = UNDEFINED):
+
+        if json is not UNDEFINED and data is not None:
+            raise ValueError("data and json parameters are mutually exclusive")
 
         self.uri = uri
         self.method = method
         self.query_string = query_string
         self.query_matcher = _create_query_matcher(self.query_string)
+        self.json = json
 
         if headers is None:
             self.headers = {}
@@ -271,6 +304,7 @@ class RequestMatcher:
             data = data.encode(data_encoding)
 
         self.data = data
+        self.data_encoding = data_encoding
 
         self.header_value_matcher = HeaderValueMatcher() if header_value_matcher is None else header_value_matcher
 
@@ -281,7 +315,8 @@ class RequestMatcher:
 
         class_name = self.__class__.__name__
         retval = "<{} ".format(class_name)
-        retval += "uri={uri!r} method={method!r} query_string={query_string!r} headers={headers!r} data={data!r}>".format_map(self.__dict__)
+        retval += "uri={uri!r} method={method!r} query_string={query_string!r} headers={headers!r} data={data!r} json={json!r}>".format_map(
+            self.__dict__)
         return retval
 
     def match_data(self, request: Request) -> bool:
@@ -296,6 +331,51 @@ class RequestMatcher:
             return True
         return request.data == self.data
 
+    def match_uri(self, request: Request) -> bool:
+        path = request.path
+
+        if isinstance(self.uri, URIPattern):
+            return self.uri.match(path)
+
+        # this is python version depending
+        # in python 3.7 and above: it is re.Pattern
+        # below python 3.7 it is _sre.SRE_Pattern which cannot be accessed directly
+        elif isinstance(self.uri, re.compile("").__class__):
+            return bool(self.uri.match(path))
+
+        else:
+            # there could be a guard isinstance(self.uri, str) been here
+            # but we want to allow any object which provides the __eq__ parameter
+            # (note: in this case it will be not typeing correct)
+            #
+            # also, python will raise TypeError when self.uri is a conflicting type
+
+            return self.uri == URI_DEFAULT or path == self.uri
+
+    def match_json(self, request: Request) -> bool:
+        """
+        Matches the request data as json.
+
+        Load the request data as json and compare it to self.json which is a
+        json-serializable data structure (eg. a dict or list).
+
+        :param request: the HTTP request
+        :return: `True` when the data is matched or no matching is required. `False` otherwise.
+        """
+        if self.json is UNDEFINED:
+            return True
+
+        try:
+            # do the decoding here as python 3.5 requires string and does not
+            # accept bytes
+            json_received = json.loads(request.data.decode(self.data_encoding))
+        except json.JSONDecodeError:
+            return False
+        except UnicodeDecodeError:
+            return False
+
+        return json_received == self.json
+
     def difference(self, request: Request) -> list:
         """
         Calculates the difference between the matcher and the request.
@@ -309,7 +389,8 @@ class RequestMatcher:
         """
 
         retval = []
-        if self.uri != URI_DEFAULT and request.path != self.uri:
+
+        if not self.match_uri(request):
             retval.append(("uri", request.path, self.uri))
 
         if self.method != METHOD_ALL and self.method != request.method:
@@ -331,6 +412,8 @@ class RequestMatcher:
         if not self.match_data(request):
             retval.append(("data", request.data, self.data))
 
+        if not self.match_json(request):
+            retval.append(("json", request.data, self.json))
         return retval
 
     def match(self, request: Request) -> bool:
@@ -470,6 +553,13 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
         :py:class:`Request` and :py:class:`Response` object which represents the
         incoming request and the outgoing response which happened during the lifetime
         of the server.
+
+    .. py:attribute:: no_handler_status_code
+
+        Attribute containing the http status code (int) which will be the response
+        status when no matcher is found for the request. By default, it is set to *500*
+        but it can be overridden to any valid http status code such as *404* if needed.
+
     """
 
     DEFAULT_LISTEN_HOST = "localhost"
@@ -498,6 +588,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             self.default_waiting_settings = WaitingSettings()
         self._waiting_settings = copy(self.default_waiting_settings)
         self._waiting_result = queue.LifoQueue(maxsize=1)
+        self.no_handler_status_code = 500
 
     def clear(self):
         """
@@ -510,6 +601,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
         self.clear_log()
         self.clear_all_handlers()
         self.permanently_failed = False
+        self.no_handler_status_code = 500
 
     def clear_assertions(self):
         """
@@ -567,14 +659,15 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
 
     def expect_request(
             self,
-            uri: str,
+            uri: Union[str, URIPattern, Pattern[str]],
             method: str = METHOD_ALL,
             data: Union[str, bytes, None] = None,
             data_encoding: str = "utf-8",
             headers: Optional[Mapping[str, str]] = None,
             query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
             header_value_matcher: Optional[HeaderValueMatcher] = None,
-            handler_type: HandlerType = HandlerType.PERMANENT) -> RequestHandler:
+            handler_type: HandlerType = HandlerType.PERMANENT,
+            json: Any = UNDEFINED) -> RequestHandler:
         """
         Create and register a request handler.
 
@@ -590,9 +683,11 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
         handlers registered, the first request must hit the first handler, and the second request must hit the second
         one, and not vice versa. If one or more ordered handler defined, those must be exhausted first.
 
-        :param uri: URI of the request. This must be an absolute path starting with ``/``.
+
+        :param uri: URI of the request. This must be an absolute path starting with ``/``, a
+            :py:class:`URIPattern` object, or a regular expression compiled by :py:func:`re.compile`.
         :param method: HTTP method of the request. If not specified (or `METHOD_ALL`
-            specified), all HTTP requests will match.
+            specified), all HTTP requests will match. Case insensitive.
         :param data: payload of the HTTP request. This could be a string (utf-8 encoded
             by default, see `data_encoding`) or a bytes object.
         :param data_encoding: the encoding used for data parameter if data is a string.
@@ -605,18 +700,24 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             object from werkzeug.
         :param header_value_matcher: :py:class:`HeaderValueMatcher` that matches values of headers.
         :param handler_type: type of handler
+        :param json: a python object (eg. a dict) whose value will be compared to the request body after it
+            is loaded as json. If load fails, this matcher will be failed also. *Content-Type* is not checked.
+            If that's desired, add it to the headers parameter.
 
         :return: Created and register :py:class:`RequestHandler`.
+
+        Parameters `json` and `data` are mutually exclusive.
         """
 
         matcher = self.create_matcher(
             uri,
-            method=method,
+            method=method.upper(),
             data=data,
             data_encoding=data_encoding,
             headers=headers,
             query_string=query_string,
             header_value_matcher=header_value_matcher,
+            json=json,
         )
         request_handler = RequestHandler(matcher)
         if handler_type == HandlerType.PERMANENT:
@@ -629,19 +730,21 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
 
     def expect_oneshot_request(
             self,
-            uri: str,
+            uri: Union[str, URIPattern, Pattern[str]],
             method: str = METHOD_ALL,
             data: Union[str, bytes, None] = None,
             data_encoding: str = "utf-8",
             headers: Optional[Mapping[str, str]] = None,
             query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-            header_value_matcher: Optional[HeaderValueMatcher] = None) -> RequestHandler:
+            header_value_matcher: Optional[HeaderValueMatcher] = None,
+            json: Any = UNDEFINED) -> RequestHandler:
         """
         Create and register a oneshot request handler.
 
         This is a method for convenience. See :py:meth:`expect_request` for documentation.
 
-        :param uri: URI of the request. This must be an absolute path starting with ``/``.
+        :param uri: URI of the request. This must be an absolute path starting with ``/``, a
+            :py:class:`URIPattern` object, or a regular expression compiled by :py:func:`re.compile`.
         :param method: HTTP method of the request. If not specified (or `METHOD_ALL`
             specified), all HTTP requests will match.
         :param data: payload of the HTTP request. This could be a string (utf-8 encoded
@@ -655,8 +758,13 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             value will be used. If multiple values needed to be handled, use ``MultiDict``
             object from werkzeug.
         :param header_value_matcher: :py:class:`HeaderValueMatcher` that matches values of headers.
+        :param json: a python object (eg. a dict) whose value will be compared to the request body after it
+            is loaded as json. If load fails, this matcher will be failed also. *Content-Type* is not checked.
+            If that's desired, add it to the headers parameter.
 
         :return: Created and register :py:class:`RequestHandler`.
+
+        Parameters `json` and `data` are mutually exclusive.
         """
 
         return self.expect_request(
@@ -668,23 +776,26 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             query_string=query_string,
             header_value_matcher=header_value_matcher,
             handler_type=HandlerType.ONESHOT,
+            json=json,
         )
 
     def expect_ordered_request(
             self,
-            uri: str,
+            uri: Union[str, URIPattern, Pattern[str]],
             method: str = METHOD_ALL,
             data: Union[str, bytes, None] = None,
             data_encoding: str = "utf-8",
             headers: Optional[Mapping[str, str]] = None,
             query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-            header_value_matcher: Optional[HeaderValueMatcher] = None) -> RequestHandler:
+            header_value_matcher: Optional[HeaderValueMatcher] = None,
+            json: Any = UNDEFINED) -> RequestHandler:
         """
         Create and register a ordered request handler.
 
         This is a method for convenience. See :py:meth:`expect_request` for documentation.
 
-        :param uri: URI of the request. This must be an absolute path starting with ``/``.
+        :param uri: URI of the request. This must be an absolute path starting with ``/``, a
+            :py:class:`URIPattern` object, or a regular expression compiled by :py:func:`re.compile`.
         :param method: HTTP method of the request. If not specified (or `METHOD_ALL`
             specified), all HTTP requests will match.
         :param data: payload of the HTTP request. This could be a string (utf-8 encoded
@@ -698,8 +809,13 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             value will be used. If multiple values needed to be handled, use ``MultiDict``
             object from werkzeug.
         :param header_value_matcher: :py:class:`HeaderValueMatcher` that matches values of headers.
+        :param json: a python object (eg. a dict) whose value will be compared to the request body after it
+            is loaded as json. If load fails, this matcher will be failed also. *Content-Type* is not checked.
+            If that's desired, add it to the headers parameter.
 
         :return: Created and register :py:class:`RequestHandler`.
+
+        Parameters `json` and `data` are mutually exclusive.
         """
 
         return self.expect_request(
@@ -711,13 +827,14 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             query_string=query_string,
             header_value_matcher=header_value_matcher,
             handler_type=HandlerType.ORDERED,
+            json=json,
         )
 
     def thread_target(self):
         """
         This method serves as a thread target when the server is started.
 
-        This should not be called directly, but can be overriden to tailor it to your needs.
+        This should not be called directly, but can be overridden to tailor it to your needs.
         """
 
         self.server.serve_forever()
@@ -737,7 +854,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
 
         If the sever is not stopped by the caller and execution reaches the end, the
         program needs to be terminated by Ctrl+C or by signal as it will not terminate until
-        the thred is stopped.
+        the thread is stopped.
 
         If the sever is already running :py:class:`HTTPServerError` will be raised. If you are
         unsure, call :py:meth:`is_running` first.
@@ -759,7 +876,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
         Notifies the server thread about the intention of the stopping, and the thread will
         terminate itself. This needs about 0.5 seconds in worst case.
 
-        Only a running server can be stopped. If the sever is not runnig, :py:class`HTTPServerError`
+        Only a running server can be stopped. If the sever is not running, :py:class`HTTPServerError`
         will be raised.
         """
         if not self.is_running():
@@ -801,7 +918,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
         This method returns a human-readable string representation of the matchers
         registered. You can observe which requests will be served, etc.
 
-        This method is primairly used when reporting errors.
+        This method is primarily used when reporting errors.
         """
         def format_handlers(handlers):
             if handlers:
@@ -833,7 +950,7 @@ class HTTPServer:   # pylint: disable=too-many-instance-attributes
             self._set_waiting_result(False)
         text = "No handler found for request {!r}.\n".format(request)
         self.add_assertion(text + self.format_matchers())
-        return Response("No handler found for this request", 500)
+        return Response("No handler found for this request", self.no_handler_status_code)
 
     def respond_permanent_failure(self):
         """
